/*
 * Function.hpp
 *
 *  Created on: Oct 9, 2017
 *      Author: julien
 */

#ifndef SRC_FUNCTION_FUNCTION_HPP_
#define SRC_FUNCTION_FUNCTION_HPP_

#include "../../OpenCV_includes.hpp"

#include <stdexcept>
#include <exception>


class Function {

	private:

		int Lookupsize; // defines the size of fak and bin_co

		vector<double> fak; // a lookup table for all factorial numbers up to "lookupsize"
							// the lookup table works like so : fak[n] = n!

		vector<vector<double>> bin_co; // lookup table for all possible binomial coefficients
									   // for n <= "lookupsize"
		// the binomial coefficient works like so : bin_co [n][k] = (n)
		//															(k)

		vector<double> si; // lookuptable for sinus; used by the sinus function in this class
		vector<double> co; // lookuptable for cosinus; used by the cosinus function in this class
		vector<double> ta; // tangens

		vector<double> asi; // arcus sinus
		vector<double> aco; // arcus cosinus
		vector<double> ata; // arcus tangens



	public:


		/**
		 * initializes an object of this class.
		 * @param int Lookupsize : defines how big the lookup tables for the factorial numbers
		 * 						   and binomial coefficient are. THis is only needed if you
		 * 						   want to translate an polynomial function in x-direction
		 * 						   without turning it into a vector function
		 */
		Function(int Lookupsize);


		/**
		 * a struct for defining a vector function
		 */
		struct vec_func{

			vector<double> poly_x; // x(t)
			vector<double> poly_y; // y(t)

			// this is the start of the defined region
			double start_t;
			// this is the end of the defined region
			double end_t;

			// the function is obviously mathematically defined outside [start_t, end_t],
			// but it makes little sense to look at that space, if it was fitted with that
			// space in mind

		};

		/**
		 * since the normal data structure "Point" only lets you use integer,
		 * this is the more precise alternative with double
		 * pPoint = precision Point
		 */
		struct pPoint{

			pPoint() : x(0), y(0){}
			pPoint(double x, double y) : x(x), y(y){}
			pPoint(Point p) : x(p.x), y(p.y) {}
			double x;
			double y;

		};



		/**
		 * evaluates a polynomial function "poly" at the point x
		 */
		static double evaluate(vector<double>* poly, double x);

		/**
		 * evaluates a vector function at the parameter t
		 */
		static pPoint evaluate_vec_func(Function::vec_func* v, double t);

		/**
		 * differantiates a polynomial function
		 */
		static vector<double> differentiate(vector<double>* poly);

		/**
		 * fits polynomial function through a set number of points.
		 * The returned function is of power "grad"
		 */
		static vector<double> square_fit(vector<pPoint> points, int grad);

		/**
		 * fits a vector function through a set number of points.
		 * both the power of x(t) and y(t) will be the size of grad
		 * points must be ordered
		 */
		static void vec_fit(vector<pPoint>* points, vec_func* out, int grad);

		/**
		 * translates a vector function f in the direction of x and y
		 */
		static void move_vec_func(vec_func* v, double delta_x, double delta_y);

		/**
		 * rotates a vector function around the zero point with the angle alpha
		 * alpha needs to be in degrees
		 */
		void rotate_vec_func(vec_func* v, double alpha);


		/**
		 * translates a polynomial function f in the direction of x and y
		 */
		vector<double> translate(vector<double> poly, double x, double y);

		/**
		 * translates a polynomial function f in the direction of y
		 */
		vector<double> move_y(vector<double> poly, double y);

		/**
		 * translates a polynomial function f in the direction of x
		 */
		vector<double> move_x(vector<double> poly, double x);

		/**
		 * Rotates a polynomial function around the zero point with the angle alpha.
		 * Returns a vector function, since a polynomial function can't be rotated and remain a
		 * normal polynomial function.
		 * After rotating a "normal" function in the likes of f(x), it needs not remain so,
		 * this is apparent when rotating a parabola
		 *
		 * @param vector<double>* poly : the polynomial function the will be rotated
		 * @param double alpha : the angle that the function will be rotated by.
		 * 						 alpha needs to be in degrees
		 */
		vec_func rotate(vector<double>* poly, double alpha);


		/**
		 * returns the sinus of an angle that is given in degrees.
		 *
		 * @param double angle : the angle for which the sinus will be calculated
		 * 						 angle needs to be in degrees!
		 * 						 Angles only matter up to one decimal place.
		 * 						 Longer angles can be provided, but the precision will not
		 * 						 improve
		 */
		double sinus(double angle);

		/**
		 * returns the cosinus of an angle that is given in degrees.
		 *
		 * @param double angle : the angle for which the cosinus will be calculated
		 * 						 angle needs to be in degrees!
		 * 						 Angles only matter up to one decimal place.
		 * 						 Longer angles can be provided, but the precision will not
		 * 						 improve
		 */
		double cosinus(double angle);


		/**
		 * returns the tangens of an angle that is given in degrees.
		 *
		 * @throws math_range_error : this error is thrown, if the provided angle is a
		 * 							  multiple of 180.
		 *
		 * @param double angle : the angle for which the tangens will be calculated
		 * 						 angle needs to be in degrees!
		 * 						 Angles only matter up to one decimal place.
		 * 						 Longer angles can be provided, but the precision will not
		 * 						 improve
		 */
		double tangens(double angle);


		/**
		 * returns the arcus sinus of a fracture that is given.
		 *
		 * @throws math_range_error : this error is thrown, if the provided fracture is
		 * 							  bigger than 1 or smaller than -1.
		 *
		 * @param double fracture : the fracture for which the arcus sinus will be calculated
		 * 						    fracture needs to be between -1 and 1!
		 * 							Fractures only matter up to two decimal places.
		 */
		double arc_sinus(double fracture);

		/**
		 * returns the arcus cosinus of a fracture that is gives.
		 *
		 * @throws math_range_error : this error is thrown, if the provided fracture is
		 * 							  bigger than 1 or smaller than -1.
		 *
		 * @param double fracture : the fracture for which the arcus cosinus will be calculated
		 * 						    fracture needs to be between -1 and 1!
		 * 							Fractures only matter up to two decimal places.
		 */
		double arc_cosinus(double fracture);


		/**
		 * returns the arcus tangens of a fracture that is gives.
		 *
		 * @param double fracture : the fracture for which the arcus tangens will be calculated
		 * 							Fractures only matter up to two decimal places.
		 */
		double arc_tangens(double fracture);


		/**
		 * uses a numeric middling algorithm (binary search) to find the t corresponding to x
		 * that is within the given precision
		 * This function is needed, as there is no clear coupling between x(t) and y(t),
		 * and this function tries to mitigate this.
		 *		 *
		 * @param Function::vec_func* v : The function in which a possible t is searched,
		 * 								  that will output x(t) = x
		 * @param double x : the x for which the value of t is searched
		 * @param unsigned double deviation : defines how precise the returned t is regarding
		 * 									  the output of x.
		 * 									  this means x(t) = x + deviation
		 * 									  The smaller the absolute value of deviation,
		 * 									  the more precise t will b, but at a cost of
		 * 									  computation time.
		 * 									  deviation needs to be positive!
		 *
		 * 	the assumpion is, that x(t_min) <= x
		 *
		 * 	returns -1 if t is not inside the defining range of v
		 */
		static double find_t(Function::vec_func* v,
							 double x,
							 double t_step_size,
							 double margin);

		/**
		 * Finds a t, so that the function length between start_t and the return value
		 * equals roughly the provided length.
		 *
		 * @param Function::vec_func* v : The function for which the t (output) is searched
		 * @param double start_t : start parameter for the given function
		 * @param double length : the length for which the t (output) is searched
		 * 						  This is the length of the function between v(start_t) and
		 * 						  v(t)
		 * @param double t_step_size : the length of the delta-t. A higher value leads to a
		 * 							   less precise output, but the calculation is a little
		 * 							   faster. (for further information look at the
		 * 							   documentation) TODO TODO
		 * @param int margin : how difference between the searched for length and the return
		 * 					   value is allowed. the more precise, the more computation power
		 * 					   is needed
		 */
		static double find_t_for_length(Function::vec_func* v,
										double start_t,
										double length,
										double t_step_size,
										double margin);

		/**
		 * Finds an approximated Length between two t's of a vector function.
		 * The method of approximation uses direct lines between t0 and t1.
		 *
		 *
		 * @param Function::vec_func* v : The function for which the length is calculated.
		 * @param double t0 : this is the parameter for the point where the length
		 * 					  measurement starts.
		 * @param double t1 : this is the parameter for the point where the length
		 * 					  measurement ends.
		 * 					  t0 < t1
		 * @param int guide_points : defines how many guide points shall be used in between
		 * 							 t0 and t1.
		 * 							 By increasing the number of guide points extra Points are
		 * 							 added between t0 and t1 for a more precise measurement.
		 * 							 for guide_points -> infinity, the length gets exact
		 */
		static double calc_length(Function::vec_func* v,
								  double t0,
								  double t1,
								  int guide_points);

		/**
		 * calculates the length of the direct line via Pythagoras
		 *
		 * @param pPoint pt1 : first point for the distance
		 * @param pPoint pt2 : second point for the distance
		 */
		static double calc_distance(pPoint pt1, pPoint pt2);

		/**
		 * calculates the curvature of the given vector lane at the point
		 * (x(t), y(t))
		 * https://de.wikipedia.org/wiki/Kr%C3%BCmmung#Ebene_Kurven
		 *
		 */
		static double calc_curvature(Function::vec_func* v, double t);


		/**
		 * converts a vector function into a collection of points.
		 * This is for a reduced representation of vector functions.
		 * For example for projecting a vector function into another 2d-planel
		 * TODO
		 * TODO
		 * TODO
		 */
		static void get_point_representation(Function::vec_func* v_src,
											 vector<Function::pPoint>* out,
											 int number_points);

		/**
		 * calculates the derivative of a vector function at a point t
		 */
		double calc_theta(Function::vec_func* v, double t);

		/**
		 * calculates the vector function that is the middle of two other
		 * vector functions. This might be needed to find the middle of lane
		 * TODO
		 */
		static void calc_middle_vec_func(Function::vec_func* v1,
										 Function::vec_func* v2,
										 Function::vec_func* out);

		/**
		 * sees if a reference (pointer) to a vector function is valid
		 * and if it is, can you actually work with the given vector function
		 */
		static bool is_usable(Function::vec_func* v);

};


// this exception is called if something is not mathematically defined
// something like arcsin(2);
class math_range_error: public exception
{
  virtual const char* what() const throw()
  {
	return "The provided value is not inside the defined range of this function";
  }
};



#endif /* SRC_FUNCTION_FUNCTION_HPP_ */
